/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_default.c 1023 2006-03-09 23:19:22Z hdaschbach $
 */


#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#endif

#include "mpl.h"
#include "mpl_crtn.h"
#include "mpl_log.h"

static int
mpl_default_probe(struct scsi_cmnd *cmd)
{
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];

	trim_to(vendor, cmd->request_buffer + 8, sizeof(vendor));
	trim_to(product, cmd->request_buffer + 16, sizeof(product));
	trim_to(rev, cmd->request_buffer + 32, sizeof(rev));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	mpl_print(MPL_INIT, MPL_INFO "Default plugin "
		  "support for: tgt %d lun %d: %s, %s, %s\n",
		  cmd->target, cmd->lun,
		  vendor, product, rev);
#else
	mpl_print(MPL_INIT, MPL_INFO "Default plugin "
		  "support for: tgt %d lun %d: %s, %s, %s\n",
		  cmd->device->id, cmd->device->lun,
		  vendor, product, rev);
#endif
	return 1;
}

static void
mpl_default_path_down(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY, MPL_INFO "Path Down Plugin %s %s\n",
		mpl_print_device(path->device), mpl_print_path(path));

	return;
}

static int
mpl_default_path_up(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY, MPL_INFO "Path Up Plugin %s %s\n",
		mpl_print_device(path->device), mpl_print_path(path));

	return 1;
}

static int
mpl_default_diagnose(struct io_stub *stub, int flag)
{
	struct mpl_path *path = stub->path;
	struct scsi_cmnd *cmd = stub->cmd;
	int rv = 0;

	mpl_print(MPL_RECOVERY, MPL_INFO "Diagnose Plugin "
		"%s %s%scmd %x result %x\n",
		mpl_print_device(path->device), mpl_print_path(path),
		(stub->done) ? " " : " internal ",
		(unsigned int)cmd->cmnd[0], cmd->result);


	/* Setup alua / active-passive on first command */
	/* Make it think we support active-active ALUA, but don't
	 * set the alua support flag in the device structure.
	 */
	path->tpgs = MPL_ALUA_ACTIVE_ACTIVE;
	path->sas  = MPL_ALUA_ACTIVE_OPT;
	path->device->active_passive = 0;  /* active - active */

	switch (flag) {
	case PLUG_INTERNAL_CMD:
		/* mpl driver ignores the return value for this case */
		break;

	case PLUG_HBEAT_CMD:
		/* mpl driver assumes a return value of 0 means the hbeat
		 * command succeeded.
		 */
		switch (cmd->cmnd[0]) {
		case INQUIRY:
			/* Setup tgt qdepth after first INQUIRY */
			path->device->tgt_queue_depth = 0; /* no tgt qdepth */
		case TEST_UNIT_READY:
			if(cmd->result)
				rv = 1; /* error */
		}
		break;

	case PLUG_SCSI_CMD:
		/* mpl driver assumes a return value of 1 means to initiate
		 * a failover and retry the command.
		 */
		break;
	}
	return rv;
}

/* In non-default plugin files the plugin table is defined as:
 * DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid)
 * instead of the table below
 */

struct mpl_plugin plugin_default
__attribute__((section("plugin.default"), aligned(sizeof(void *)))) =
{
	"default",
	mpl_default_probe,
	NULL,
	mpl_default_diagnose,
	mpl_default_path_down,
	mpl_default_path_up,
	NULL
};
